package com.virjar.g4proxyserver.netty.handler;


import com.virjar.g4proxyserver.netty.server.ProxyChannelManager;
import com.virjar.nat.protocol.Constants;
import com.virjar.nat.protocol.ProxyMessage;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.Channel;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelOption;
import io.netty.channel.SimpleChannelInboundHandler;
import lombok.extern.slf4j.Slf4j;


@Slf4j
@ChannelHandler.Sharable
public class NatServerChannelHandler extends SimpleChannelInboundHandler<ProxyMessage> {


    @Override
    protected void channelRead0(ChannelHandlerContext ctx, ProxyMessage proxyMessage) throws Exception {
        log.info("recieved proxy message, type is:{}", proxyMessage.getTypeReadable());
        switch (proxyMessage.getType()) {
            case ProxyMessage.TYPE_HEARTBEAT:
                handleHeartbeatMessage(ctx, proxyMessage);
                break;
            case ProxyMessage.C_TYPE_AUTH:
                handleAuthMessage(ctx, proxyMessage);
                break;
            case ProxyMessage.TYPE_CONNECT:
                handleConnectMessage(ctx, proxyMessage);
                break;
            case ProxyMessage.TYPE_DISCONNECT:
                handleDisconnectMessage(ctx, proxyMessage);
                break;
            case ProxyMessage.P_TYPE_TRANSFER:
                handleTransferMessage(ctx, proxyMessage);
                break;
            default:
                break;
        }
    }

    /**
     * 心跳处理
     *
     * @param ctx
     * @param proxyMessage
     */
    private void handleHeartbeatMessage(ChannelHandlerContext ctx, ProxyMessage proxyMessage) {
        ProxyMessage heartbeatMessage = new ProxyMessage();
        heartbeatMessage.setSerialNumber(heartbeatMessage.getSerialNumber());
        heartbeatMessage.setType(ProxyMessage.TYPE_HEARTBEAT);
        log.info("response heartbeat message {}", ctx.channel());
        ctx.channel().writeAndFlush(heartbeatMessage);
    }

    /**
     * 认证消息，检测clientKey是否正确
     *
     * @param ctx
     * @param proxyMessage
     */
    private void handleAuthMessage(ChannelHandlerContext ctx, ProxyMessage proxyMessage) {
        String clientKey = proxyMessage.getUri();
        log.info("client connect :{}", clientKey);

        ProxyChannelManager.addCmdChannel(clientKey, ctx.channel());
    }

    /**
     * 代理后端服务器建立连接消息
     *
     * @param ctx
     * @param proxyMessage
     */
    private void handleConnectMessage(ChannelHandlerContext ctx, ProxyMessage proxyMessage) {
        String uri = proxyMessage.getUri();
        if (uri == null) {
            ctx.channel().close();
            log.warn("ConnectMessage:null uri");
            return;
        }

        //userId + "@" + keys
        String[] tokens = uri.split("@");
        if (tokens.length != 2) {
            ctx.channel().close();
            log.warn("ConnectMessage:error uri");
            return;
        }

        Channel cmdChannel = ProxyChannelManager.getCmdChannel(tokens[1]);
        if (cmdChannel == null) {
            ctx.channel().close();
            log.warn("ConnectMessage:error cmd channel key {}", tokens[1]);
            return;
        }

        //user mapping channel,根据user Id可以查询到，是真实用户到来的时候，创建并绑定的
        Channel userMappingChannel = ProxyChannelManager.getUserChannel(cmdChannel, tokens[0]);
        if (userMappingChannel != null) {
            //ctx.channel() 是natServer->natClient的链接通道，为client在收到connect命令之后，主动和natServer建立的数据通道
            Channel natDataChannel = ctx.channel();
            natDataChannel.attr(Constants.USER_ID).set(tokens[0]);
            natDataChannel.attr(Constants.CLIENT_KEY).set(tokens[1]);

            //双向关联
            natDataChannel.attr(Constants.NEXT_CHANNEL).set(userMappingChannel);
            userMappingChannel.attr(Constants.NEXT_CHANNEL).set(natDataChannel);
            // 代理客户端与后端服务器连接成功，修改用户连接为可读状态，之后触发UserMappingChannelHandler channelRead0，读取外部真正的数据
            userMappingChannel.config().setOption(ChannelOption.AUTO_READ, true);
        }
    }

    /**
     * 代理数据传输
     *
     * @param ctx
     * @param proxyMessage
     */
    private void handleTransferMessage(ChannelHandlerContext ctx, ProxyMessage proxyMessage) {
        Channel userMappingChannel = ctx.channel().attr(Constants.NEXT_CHANNEL).get();
        if (userMappingChannel != null) {
            ByteBuf buf = ctx.alloc().buffer(proxyMessage.getData().length);
            buf.writeBytes(proxyMessage.getData());
            userMappingChannel.writeAndFlush(buf);
        }
    }

    /**
     * 代理后端服务器断开连接消息
     *
     * @param ctx
     * @param proxyMessage
     */
    private void handleDisconnectMessage(ChannelHandlerContext ctx, ProxyMessage proxyMessage) {
        String clientKey = ctx.channel().attr(Constants.CLIENT_KEY).get();

        // 代理连接没有连上服务器由控制连接发送用户端断开连接消息
        if (clientKey == null) {
            String userId = proxyMessage.getUri();
            Channel userChannel = ProxyChannelManager.removeUserChannelFromCmdChannel(ctx.channel(), userId);
            if (userChannel != null) {
                // 数据发送完成后再关闭连接，解决http1.0数据传输问题
                userChannel.writeAndFlush(Unpooled.EMPTY_BUFFER).addListener(ChannelFutureListener.CLOSE);
            }
            return;
        }

        Channel cmdChannel = ProxyChannelManager.getCmdChannel(clientKey);
        if (cmdChannel == null) {
            log.warn("ConnectMessage:error cmd channel key {}", ctx.channel().attr(Constants.CLIENT_KEY).get());
            return;
        }

        Channel userChannel = ProxyChannelManager.removeUserChannelFromCmdChannel(cmdChannel, ctx.channel().attr(Constants.USER_ID).get());
        if (userChannel != null) {
            // 数据发送完成后再关闭连接，解决http1.0数据传输问题
            userChannel.writeAndFlush(Unpooled.EMPTY_BUFFER).addListener(ChannelFutureListener.CLOSE);
            ctx.channel().attr(Constants.NEXT_CHANNEL).remove();
            ctx.channel().attr(Constants.CLIENT_KEY).remove();
            ctx.channel().attr(Constants.USER_ID).remove();
        }
    }


}
