package com.virjar.g4proxyserver.netty.server;

import com.google.common.collect.Maps;
import com.virjar.g4proxyserver.netty.AvailablePortManager;
import com.virjar.g4proxyserver.netty.handler.NatServerChannelHandler;
import com.virjar.g4proxyserver.netty.handler.ServerIdleCheckHandler;
import com.virjar.g4proxyserver.netty.handler.UserMappingChannelHandler;
import com.virjar.nat.protocol.Constants;
import com.virjar.nat.protocol.ProxyMessageDecoder;
import com.virjar.nat.protocol.ProxyMessageEncoder;

import java.util.Collection;
import java.util.Map;

import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import lombok.extern.slf4j.Slf4j;

@Slf4j
public class ProxyServer {

    /**
     * max packet is 2M.
     */
    private static final int MAX_FRAME_LENGTH = 2 * 1024 * 1024;

    private static final int LENGTH_FIELD_OFFSET = 0;

    private static final int LENGTH_FIELD_LENGTH = 4;

    private static final int INITIAL_BYTES_TO_STRIP = 0;

    private static final int LENGTH_ADJUSTMENT = 0;

    private ServerBootstrap natServerBootStrap;
    private ServerBootstrap userMapServerBootStrap;

    private Map<Integer, ChannelFuture> portMappingChannelMap = Maps.newConcurrentMap();


    private static ProxyServer instance = new ProxyServer();

    private ProxyServer() {
    }

    public static ProxyServer getInstance() {
        return instance;
    }

    public Collection<Integer> mappingPort() {
        return portMappingChannelMap.keySet();
    }

    public boolean openMappingPort(int port) {
        ChannelFuture channelFuture = userMapServerBootStrap.bind(port);
        try {
            channelFuture.get();
        } catch (Exception e) {
            log.error("wait for port binding error", e);
            return false;
        }
        if (!channelFuture.isSuccess()) {
            log.warn("bind mapping port:{} failed", port);
            return false;
        }
        portMappingChannelMap.put(port, channelFuture);
        return true;
    }

    public void closeMappingPort(int port) {
        ChannelFuture channelFuture = portMappingChannelMap.get(port);
        if (channelFuture != null) {
            channelFuture.channel().close();
        }
        portMappingChannelMap.remove(port);
    }

    public void start() {
        natServerBootStrap = new ServerBootstrap();
        NioEventLoopGroup serverBossGroup = new NioEventLoopGroup();
        NioEventLoopGroup serverWorkerGroup = new NioEventLoopGroup();
        natServerBootStrap.group(serverBossGroup, serverWorkerGroup).channel(NioServerSocketChannel.class).childHandler(new ChannelInitializer<SocketChannel>() {
            @Override
            public void initChannel(SocketChannel ch) throws Exception {
                ch.pipeline().addLast(new ProxyMessageDecoder(MAX_FRAME_LENGTH, LENGTH_FIELD_OFFSET, LENGTH_FIELD_LENGTH, LENGTH_ADJUSTMENT, INITIAL_BYTES_TO_STRIP));
                ch.pipeline().addLast(new ProxyMessageEncoder());
                ch.pipeline().addLast(new ServerIdleCheckHandler());
                ch.pipeline().addLast(new NatServerChannelHandler());
            }
        });
        log.info("start netty [Proxy Server] server ,port:" + Constants.g4ProxyServerPort);
        natServerBootStrap.bind(Constants.g4ProxyServerPort);

        userMapServerBootStrap = new ServerBootstrap();
        serverBossGroup = new NioEventLoopGroup();
        serverWorkerGroup = new NioEventLoopGroup();
        userMapServerBootStrap.group(serverBossGroup, serverWorkerGroup).channel(NioServerSocketChannel.class).childHandler(new ChannelInitializer<SocketChannel>() {
            @Override
            public void initChannel(SocketChannel ch) throws Exception {
                ch.pipeline().addLast(new UserMappingChannelHandler());
            }
        });

    }

}
